"""Base module variables."""
import importlib.util
import json
import os.path as op
from pathlib import Path

# Get version
spec = importlib.util.spec_from_file_location(
    "_version", op.join(op.dirname(__file__), "ddmra/_version.py")
)
_version = importlib.util.module_from_spec(spec)
spec.loader.exec_module(_version)

VERSION = _version.get_versions()["version"]
del _version

# Get list of authors from Zenodo file
with open(op.join(op.dirname(__file__), ".zenodo.json"), "r") as fo:
    zenodo_info = json.load(fo)
authors = [author["name"] for author in zenodo_info["creators"]]
author_names = []
for author in authors:
    if ", " in author:
        author_names.append(author.split(", ")[1] + " " + author.split(", ")[0])
    else:
        author_names.append(author)

# Get package description from README
# Since this file is executed from ../setup.py, the path to the README is determined by the
# location of setup.py.
readme_path = Path(__file__).parent.joinpath("README.md")
longdesc = readme_path.open().read()

# Fields
AUTHOR = "Taylor Salo"
COPYRIGHT = "Copyright 2021, Taylor Salo"
CREDITS = author_names
LICENSE = "LGPL 2.1"
MAINTAINER = "Taylor Salo"
EMAIL = "tsalo006@fiu.edu"
STATUS = "Prototype"
URL = "https://github.com/ME-ICA/ddmra"
PACKAGENAME = "ddmra"
DESCRIPTION = "A Python package for distance-dependent motion-related artifact (DDMRA) analysis."
LONGDESC = longdesc

DOWNLOAD_URL = "https://github.com/tsalo/{name}/archive/{ver}.tar.gz".format(
    name=PACKAGENAME, ver=VERSION
)

REQUIRES = [
    "joblib",
    "nibabel",
    "nilearn",
    "numpy>=1.15",
    "scipy>=1.3.3",
    "seaborn",
]

TESTS_REQUIRES = [
    "codecov",
    "coverage<5.0",
    "flake8>=3.7",
    "flake8-black",
    "flake8-isort",
    "pytest",
    "pytest-cov",
]

EXTRA_REQUIRES = {
    "dev": ["versioneer"],
    "doc": [
        "nbsphinx",
        "sphinx>=1.5.3",
        "sphinx_rtd_theme",
    ],
    "tests": TESTS_REQUIRES,
    "duecredit": ["duecredit"],
}

ENTRY_POINTS = {}

# Enable a handle to install all extra dependencies at once
EXTRA_REQUIRES["all"] = list(set([v for deps in EXTRA_REQUIRES.values() for v in deps]))

# Supported Python versions using PEP 440 version specifiers
# Should match the same set of Python versions as classifiers
PYTHON_REQUIRES = ">=3.6"

# Package classifiers
CLASSIFIERS = [
    "Development Status :: 2 - Pre-Alpha",
    "Intended Audience :: Science/Research",
    "Topic :: Scientific/Engineering :: Information Analysis",
    "License :: OSI Approved :: GNU Library or Lesser General Public License (LGPL)",
    "Programming Language :: Python :: 3.6",
    "Programming Language :: Python :: 3.7",
    "Programming Language :: Python :: 3.8",
    "Programming Language :: Python :: 3.9",
]



"""Tests for the ddmra.utils module."""
import math

import numpy as np

from ddmra.utils import null_to_p


def test_null_to_p_float():
    """Test null_to_p with single float input, assuming asymmetric null dist."""
    null = [-10, -9, -9, -3, -2, -1, -1, 0, 1, 1, 1, 2, 3, 3, 4, 4, 7, 8, 8, 9]

    # Two-tailed
    assert math.isclose(null_to_p(0, null, "two"), 0.8)
    assert math.isclose(null_to_p(9, null, "two"), 0.1)
    assert math.isclose(null_to_p(10, null, "two"), 0.05)
    assert math.isclose(null_to_p(-9, null, "two"), 0.3)
    assert math.isclose(null_to_p(-10, null, "two"), 0.1)
    # Still 0.05 because minimum valid p-value is 1 / len(null)
    result = null_to_p(20, null, "two")
    assert result == null_to_p(-20, null, "two")
    assert math.isclose(result, 0.05)

    # Left/lower-tailed
    assert math.isclose(null_to_p(9, null, "lower"), 0.95)
    assert math.isclose(null_to_p(-9, null, "lower"), 0.15)
    assert math.isclose(null_to_p(0, null, "lower"), 0.4)

    # Right/upper-tailed
    assert math.isclose(null_to_p(9, null, "upper"), 0.05)
    assert math.isclose(null_to_p(-9, null, "upper"), 0.95)
    assert math.isclose(null_to_p(0, null, "upper"), 0.65)

    # Test that 1/n(null) is preserved with extreme values
    nulldist = np.random.normal(size=10000)
    assert math.isclose(null_to_p(20, nulldist, "two"), 1 / 10000)
    assert math.isclose(null_to_p(20, nulldist, "lower"), 1 - 1 / 10000)


def test_null_to_p_float_symmetric():
    """Test null_to_p with single float input, assuming symmetric null dist."""
    null = [-10, -9, -9, -3, -2, -1, -1, 0, 1, 1, 1, 2, 3, 3, 4, 4, 7, 8, 8, 9]

    # Only need to test two-tailed; symmetry is irrelevant for one-tailed
    assert math.isclose(null_to_p(0, null, "two", symmetric=True), 0.95)
    result = null_to_p(9, null, "two", symmetric=True)
    assert result == null_to_p(-9, null, "two", symmetric=True)
    assert math.isclose(result, 0.2)
    result = null_to_p(10, null, "two", symmetric=True)
    assert result == null_to_p(-10, null, "two", symmetric=True)
    assert math.isclose(result, 0.05)
    # Still 0.05 because minimum valid p-value is 1 / len(null)
    result = null_to_p(20, null, "two", symmetric=True)
    assert result == null_to_p(-20, null, "two", symmetric=True)
    assert math.isclose(result, 0.05)


def test_null_to_p_array():
    """Test nimare.stats.null_to_p with 1d array input."""
    N = 10000
    nulldist = np.random.normal(size=N)
    t = np.sort(np.random.normal(size=N))
    p = np.sort(null_to_p(t, nulldist))
    assert p.shape == (N,)
    assert (p < 1).all()
    assert (p > 0).all()
    # Resulting distribution should be roughly uniform
    assert np.abs(p.mean() - 0.5) < 0.02
    assert np.abs(p.var() - 1 / 12) < 0.02



"""Generate distance-dependent motion-related artifact plots.

The rank for the intercept (smoothing curve at 35mm) indexes general dependence
on motion (i.e., a mix of global and focal effects), while the rank for the
slope (difference in smoothing curve at 100mm and 35mm) indexes distance
dependence (i.e., focal effects).
"""
import os.path as op

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns

from .utils import assess_significance

sns.set_style("white")


def plot_analysis(
    data_points,
    distances,
    smoothing_curve,
    curve_distances,
    null_smoothing_curves,
    n_lines=50,
    metric_name=None,
    ylim=(-0.5, 0.5),
    fig=None,
    ax=None,
):
    """Generate plot for a DDMRA analysis.

    Parameters
    ----------
    data_points : numpy.ndarray of shape (n_edges,)
        DDMRA-metric values for all unique ROI-to-ROI edges, not including self-self edges.
    distances : numpy.ndarray of shape (n_edges,)
        All distances associated with data in data_points, in mm.
    smoothing_curve : numpy.ndarray of shape (n_unique_edge_distances,)
        Smoothing curve of data points, produced using a moving average.
    curve_distances : numpy.ndarray of shape (n_unique_edge_distances,)
        Edge distances, in mm, of data points in smoothing curve.
    null_smoothing_curves : numpy.ndarray of shape (n_iters, n_unique_edge_distances)
        Smoothing curves for all permutations, to be used as a null distribution.
    n_lines : int, optional
        Number of null smoothing curves to plot in the figure. Default is 50.
    metric_name : None or str, optional
        Label for the Y-axis of the figure, indicating the analysis' metric's name.
    ylim : tuple, optional
        Y-limits.
    fig : None or matplotlib.pyplot.Figure, optional
        Figure object.
    ax : None or matplotlib.pyplot.Axes, optional
        Axes object.

    Returns
    -------
    fig : matplotlib.pyplot.Figure
        Updated Figure object.
    ax : matplotlib.pyplot.Axes
        Updated Axes object.
    """
    assert data_points.ndim == distances.ndim == smoothing_curve.ndim == curve_distances.ndim == 1
    assert null_smoothing_curves.ndim == 2
    assert data_points.shape == distances.shape
    assert smoothing_curve.shape[0] == curve_distances.shape[0] == null_smoothing_curves.shape[1]

    if not ax:
        fig, ax = plt.subplots(figsize=(16, 10))

    V1, V2 = 35, 100  # distances to evaluate
    n_lines = np.minimum(n_lines, null_smoothing_curves.shape[0])

    p_intercept, p_slope = assess_significance(
        smoothing_curve,
        null_smoothing_curves,
        curve_distances,
        V1,
        V2,
    )

    sns.regplot(
        x=distances,
        y=data_points,
        ax=ax,
        scatter=True,
        fit_reg=False,
        scatter_kws={"color": "red", "s": 5.0, "alpha": 1},
    )
    ax.axhline(0, xmin=0, xmax=np.max(distances) + 100, color="black", linewidth=3)

    for i_line in range(n_lines):
        ax.plot(curve_distances, null_smoothing_curves[i_line, :], color="black")

    ax.plot(curve_distances, smoothing_curve, color="white")

    ax.set_ylabel(metric_name, fontsize=32, labelpad=-30)
    ax.set_yticks(ylim)
    ax.set_yticklabels(ylim, fontsize=32)
    ax.set_ylim(ylim)

    ax.set_xlabel("Distance (mm)", fontsize=32)
    ax.set_xticks([0, 50, 100, 150])
    ax.set_xticklabels([])
    ax.set_xlim(0, np.ceil(np.max(distances) / 10) * 10)

    ax.annotate(
        f"{V1} mm: {p_intercept:.04f}\n{V1}-{V2} mm: {p_slope:.04f}",
        xy=(1, 0),
        xycoords="axes fraction",
        xytext=(-20, 20),
        textcoords="offset pixels",
        horizontalalignment="right",
        verticalalignment="bottom",
        fontsize=32,
    )
    fig.tight_layout()

    return fig, ax


def plot_results(in_dir):
    """Plot the results for all three analyses from a workflow run and save to a file.

    This function leverages the output file structure of :func:`workflows.run_analyses`.
    It writes out an image (analysis_results.png) to the output directory.

    Parameters
    ----------
    in_dir : str
        Path to the output directory of a ``run_analyses`` run.
    """
    METRIC_LABELS = {
        "qcrsfc": r"QC:RSFC $z_{r}$" + "\n(QC = mean FD)",
        "highlow": "High-low motion\n" + r"${\Delta}z_{r}$",
        "scrubbing": "Scrubbing\n" + r"${\Delta}z_{r}$",
    }
    YLIMS = {
        "qcrsfc": (-1.0, 1.0),
        "highlow": (-1.0, 1.0),
        "scrubbing": (-0.05, 0.05),
    }
    analysis_values = pd.read_table(op.join(in_dir, "analysis_values.tsv.gz"))
    smoothing_curves = pd.read_table(op.join(in_dir, "smoothing_curves.tsv.gz"))
    null_curves = np.load(op.join(in_dir, "null_smoothing_curves.npz"))

    fig, axes = plt.subplots(figsize=(8, 24), nrows=len(METRIC_LABELS))

    for i_analysis, (analysis_type, label) in enumerate(METRIC_LABELS.items()):
        values = analysis_values[analysis_type].values
        smoothing_curve = smoothing_curves[analysis_type].values

        fig, axes[i_analysis] = plot_analysis(
            values,
            analysis_values["distance"],
            smoothing_curve,
            smoothing_curves["distance"],
            null_curves[analysis_type],
            n_lines=50,
            ylim=YLIMS[analysis_type],
            metric_name=label,
            fig=fig,
            ax=axes[i_analysis],
        )

    fig.savefig(op.join(in_dir, "analysis_results.png"), dpi=100)



"""Miscellaneous utility functions for the DDMRA package."""
import numpy as np


def get_val(x_arr, y_arr, x_val):
    """Perform interpolation to get value from y_arr at index of x_val based on x_arr.

    TODO: Better documentation

    Parameters
    ----------
    x_arr : (N,) numpy.ndarray

    y_arr : ([X], N) numpy.ndarray

    x_val : float
        Position in x_arr for which to estimate value in y_arr.

    Returns
    -------
    y_val : float or (X,) numpy.ndarray
        Value in y_arr corresponding to position x_val in x_arr.
    """
    if y_arr.ndim == 2:
        y_arr = y_arr.T

    assert x_arr.shape[0] == y_arr.shape[0]

    temp_idx = np.where(x_arr == x_val)[0]
    if len(temp_idx) > 0:
        if y_arr.ndim == 2:
            y_val = np.mean(y_arr[temp_idx, :], axis=0)
        else:
            y_val = np.mean(y_arr[temp_idx])

    else:
        val1 = x_arr[np.where(x_arr <= x_val)[0][-1]]
        val2 = x_arr[np.where(x_arr >= x_val)[0][0]]
        temp_idx1 = np.where(x_arr == val1)[0]
        temp_idx2 = np.where(x_arr == val2)[0]
        temp_idx = np.unique(np.hstack((temp_idx1, temp_idx2)))
        if y_arr.ndim == 2:
            y_val = np.zeros(y_arr.shape[1])
            for i in range(y_val.shape[0]):
                temp_y_arr = y_arr[:, i]
                y_val[i] = np.interp(xp=x_arr[temp_idx], fp=temp_y_arr[temp_idx], x=x_val)
        else:
            y_val = np.interp(xp=x_arr[temp_idx], fp=y_arr[temp_idx], x=x_val)

    return y_val


def null_to_p(test_value, null_array, tail="two", symmetric=False):
    """Return p-value for test value(s) against null array.

    Parameters
    ----------
    test_value : 1D array_like
        Values for which to determine p-value.
    null_array : 1D array_like
        Null distribution against which test_value is compared.
    tail : {'two', 'upper', 'lower'}, optional
        Whether to compare value against null distribution in a two-sided
        ('two') or one-sided ('upper' or 'lower') manner.
        If 'upper', then higher values for the test_value are more significant.
        If 'lower', then lower values for the test_value are more significant.
        Default is 'two'.
    symmetric : bool
        When tail="two", indicates how to compute p-values. When False (default),
        both one-tailed p-values are computed, and the two-tailed p is double
        the minimum one-tailed p. When True, it is assumed that the null
        distribution is zero-centered and symmetric, and the two-tailed p-value
        is computed as P(abs(test_value) >= abs(null_array)).

    Returns
    -------
    p_value : :obj:`float`
        P-value(s) associated with the test value when compared against the null
        distribution. Return type matches input type (i.e., a float if
        test_value is a single float, and an array if test_value is an array).

    Notes
    -----
    P-values are clipped based on the number of elements in the null array.
    Therefore no p-values of 0 or 1 should be produced.
    When the null distribution is known to be symmetric and centered on zero,
    and two-tailed p-values are desired, use symmetric=True, as it is
    approximately twice as efficient computationally, and has lower variance.
    """
    if tail not in {"two", "upper", "lower"}:
        raise ValueError('Argument "tail" must be one of ["two", "upper", "lower"]')

    return_first = isinstance(test_value, (float, int))
    test_value = np.atleast_1d(test_value)
    null_array = np.array(null_array)

    # For efficiency's sake, if there are more than 1000 values, pass only the unique
    # values through percentileofscore(), and then reconstruct.
    if len(test_value) > 1000:
        reconstruct = True
        test_value, uniq_idx = np.unique(test_value, return_inverse=True)
    else:
        reconstruct = False

    def compute_p(t, null):
        null = np.sort(null)
        idx = np.searchsorted(null, t, side="left").astype(float)
        return 1 - idx / len(null)

    if tail == "two":
        if symmetric:
            p = compute_p(np.abs(test_value), np.abs(null_array))
        else:
            p_l = compute_p(test_value, null_array)
            p_r = compute_p(test_value * -1, null_array * -1)
            p = 2 * np.minimum(p_l, p_r)
    elif tail == "lower":
        p = compute_p(test_value * -1, null_array * -1)
    else:
        p = compute_p(test_value, null_array)

    # ensure p_value in the following range:
    # smallest_value <= p_value <= (1.0 - smallest_value)
    smallest_value = np.maximum(np.finfo(float).eps, 1.0 / len(null_array))
    result = np.maximum(smallest_value, np.minimum(p, 1.0 - smallest_value))

    if reconstruct:
        result = result[uniq_idx]

    return result[0] if return_first else result


def fast_pearson(X, y):
    """Fast correlations between y and each row of X, for QC-RSFC.

    Checked for accuracy. From http://qr.ae/TU1B9C

    Parameters
    ----------
    X : (i, j) numpy.ndarray
        Matrix for which each row is correlated with y.
    y : (j,) numpy.ndarray
        Array with which to correlate each row of X.

    Returns
    -------
    pearsons : (i,) numpy.ndarray
        Row-wise correlations between X and y.
    """
    assert len(X.shape) == 2
    assert len(y.shape) == 1
    assert y.shape[0] == X.shape[1]
    y_bar = np.mean(y)
    y_intermediate = y - y_bar
    X_bar = np.mean(X, axis=1)[:, np.newaxis]
    X_intermediate = X - X_bar
    nums = X_intermediate.dot(y_intermediate)
    y_sq = np.sum(np.square(y_intermediate))
    X_sqs = np.sum(np.square(X_intermediate), axis=1)
    denoms = np.sqrt(y_sq * X_sqs)
    pearsons = nums / denoms
    return pearsons


def get_fd_power(motion, order=["x", "y", "z", "r", "p", "ya"], unit="deg", radius=50):
    """Calculate Framewise Displacement (Power version).

    This is meant to be very flexible, but is really only here for the test ADHD dataset.

    Parameters
    ----------
    motion : (T, 6) numpy.ndarray
        Translation and rotation motion parameters.
    order : (6,) list
        Order of the motion parameters in motion array. Some combination of
        'x', 'y', 'z', 'r', 'p', and 'ya'. Default has translations then
        rotations.
    unit : {'deg', 'rad'}
        Which unit the rotation parameters are in. 'deg' for degree (default in
        SPM and AfNI), 'rad' for radian (default in FSL). Default is 'deg'.
    radius : float or int
        Radius of brain in millimeters. Default is 50, as used by Power.

    Returns
    -------
    fd : (T,) numpy.ndarray
        Framewise displacement values.
    """
    des_order = ["x", "y", "z", "r", "p", "ya"]
    reorder = [order.index(i) for i in des_order]
    motion_reordered = motion[:, reorder]
    if unit == "deg":
        motion_reordered[:, 3:] = motion_reordered[:, 3:] * radius * (np.pi / 180.0)
    elif unit == "rad":
        motion_reordered[:, 3:] = motion_reordered[:, 3:] * radius
    else:
        raise Exception("Rotation units must be degrees or radians.")

    deriv = np.vstack((np.array([[0, 0, 0, 0, 0, 0]]), np.diff(motion_reordered, axis=0)))
    fd = np.sum(np.abs(deriv), axis=1)
    return fd


def moving_average(values, window):
    """Calculate running average along values array.

    Parameters
    ----------
    values : (N,) numpy.ndarray
        Values over which to compute average.
    window : int
        Sliding window size over which to average values.

    Returns
    -------
    sma : (N,) numpy.ndarray
        Each value in sma is the average of the `window` values in values
        surrounding that position. So sma[1000] (with a window of 500) will be
        the mean of values[750:1250]. Positions at the beginning and end will
        be NaNs.
    """
    assert window % 2 == 0, "Window must be divisible by 2."

    weights = np.ones(window) / window
    sma = np.convolve(values, weights, "valid")
    buffer = np.full(shape=int(window / 2), fill_value=np.nan)
    sma = np.hstack((buffer, sma, buffer))[:-1]

    return sma


def average_across_distances(values, distances):
    nonnan_idx = np.where(~np.isnan(values))[0]
    unique_retained_distances, retained_distance_idxs = np.unique(
        distances[nonnan_idx],
        return_inverse=True,
    )
    distance_val_sums = np.bincount(retained_distance_idxs, values[nonnan_idx])
    distance_val_means = distance_val_sums / np.bincount(retained_distance_idxs)
    return distance_val_means, unique_retained_distances


def assess_significance(curve, null_curves, distances, intercept_distance, v2):
    """Measure significance of smoothing curve against null curves.

    Parameters
    ----------
    curve : numpy.ndarray of shape (U,)
        U is distance value.
    null_curves : numpy.ndarray of shape (P, U)
        U is distance value, P is permutation.
    distances : numpy.ndarray of shape (U,)
        Distances in mm.
    intercept_distance : float
    v2 : float

    Returns
    -------
    p_inter : float
        P-value for intercept. Higher intercepts with lower p-values
        indicate both focal and global effects of the QC metric on functional connectivity.
    p_slope : float
        P-value for slope.
        Slope, in this case, is calculated as the intercept's value minus the second location's
        value, so that negative slopes go *up* and positive slopes go *down*.
        Higher slopes (i.e., those going down) with lower p-values
        indicate focal effects of the QC metric on functional connectivity
        (i.e., enhanced local connectivity and decrease long-distance connectivity due to the
        QC measure).
    """
    assert curve.ndim == distances.ndim == 1
    assert null_curves.ndim == 2
    assert curve.shape[0] == null_curves.shape[1] == distances.shape[0]

    intercept = get_val(distances, curve, intercept_distance)
    slope = intercept - get_val(distances, curve, v2)
    perm_intercepts = get_val(distances, null_curves, intercept_distance)
    perm_slopes = perm_intercepts - get_val(distances, null_curves, v2)

    p_inter = null_to_p(intercept, perm_intercepts, tail="upper")
    p_slope = null_to_p(slope, perm_slopes, tail="upper")
    return p_inter, p_slope



"""DDMRA: A Python package for distance-dependent motion-related artifact analysis."""

from ._version import get_versions
from .analysis import highlow_analysis, qcrsfc_analysis, scrubbing_analysis
from .plotting import plot_analysis, plot_results
from .workflows import run_analyses

__version__ = get_versions()["version"]

__all__ = [
    "__version__",
    "highlow_analysis",
    "qcrsfc_analysis",
    "scrubbing_analysis",
    "run_analyses",
    "plot_analysis",
    "plot_results",
]



"""Perform distance-dependent motion-related artifact analyses."""
import logging

import numpy as np
from joblib import Parallel, delayed

from .utils import average_across_distances, fast_pearson, moving_average

LGR = logging.getLogger("analysis")


def scrubbing_analysis(qc_values, group_timeseries, edge_sorting_idx, qc_thresh=0.2, perm=True):
    """Perform Power scrubbing analysis.

    Note that correlations from scrubbed timeseries are subtracted from correlations from
    unscrubbed timeseries, which is the opposite to the original Power method.
    This reverses the signs of the results, but makes inference similar to that of the
    QCRSFC and high-low motion analyses.

    Parameters
    ----------
    qc_values : (N,) list
        List of (T,) arrays
    group_timeseries : (N,) list
        List of (T, R) arrays
    edge_sorting_idx : numpy.ndarray of shape (n_edges,)
        Sorting index for upper triangle (not including self-self edges) of correlation matrix.
        This will sort the 1D array by ascending physical distance of the ROI-ROI pairs.
    qc_thresh : float
        Threshold to apply to QC values for identifying bad volumes.
    perm : bool
        Whether the call is part of the null distribution permutations or for the real deal.

    Returns
    -------
    mean_delta_r : numpy.ndarray
        Average (across subjects) change in correlation coefficient from
        unscrubbed to scrubbed timeseries for each pair of ROIs. Length of
        array will be ((R * R) - R) / 2 (upper triangle of RxR correlation
        matrix).

    Notes
    -----
    The basic process for the scrubbing analysis is:
    1. Exclude any subjects with more than 50% excluded volumes or 0% excluded volumes.
    2. For each subject, correlate each ROI's time series with every other ROI's
       time series to produce standard correlation matrix.
    3. Apply QC metric threshold to "scrub" (i.e., remove bad volumes) time series,
       then compute scrubbing correlation matrix.
    4. Select the upper triangle (minus the diagonal) from the standard and scrubbing
       correlation matrices, flattening them both to 1D.
    5. Fisher's z-transform all correlation coefficients from both the standard and scrubbing
       vectors. **WARNING** The Power et al. version does not do this.
    6. Subtract the scrubbing z-values from the standard z-values.
       **WARNING** This is the opposite of how Power et al. did this!
    7. Average the difference values across participants.
    """
    n_rois = group_timeseries[0].shape[0]
    triu_idx = np.triu_indices(n_rois, k=1)
    n_pairs = len(triu_idx[0])
    n_subjects = len(group_timeseries)
    delta_zs = np.zeros((n_subjects, n_pairs))
    c = 0  # included subject counter
    for i_subj in range(n_subjects):
        ts_arr = group_timeseries[i_subj]
        qc_arr = qc_values[i_subj]
        keep_idx = qc_arr <= qc_thresh

        # Subjects with no timepoints excluded or with more than 50% excluded
        # will be excluded from analysis.
        prop_incl = np.sum(keep_idx) / qc_arr.shape[0]
        if (prop_incl >= 0.5) and (prop_incl != 1.0):
            scrubbed_ts = ts_arr[:, keep_idx]
            raw_corrs = np.corrcoef(ts_arr)
            raw_corrs = raw_corrs[triu_idx]
            raw_zs = np.arctanh(raw_corrs)
            scrubbed_corrs = np.corrcoef(scrubbed_ts)
            scrubbed_corrs = scrubbed_corrs[triu_idx]
            scrubbed_zs = np.arctanh(scrubbed_corrs)
            delta_zs[c, :] = raw_zs - scrubbed_zs  # opposite of Power
            c += 1

    if not perm:
        LGR.info(f"{c} of {n_subjects} subjects retained in scrubbing analysis")

    # Remove extra rows corresponding to bad subjects
    delta_zs = delta_zs[:c, :]
    # Average over subjects
    mean_delta_z = np.mean(delta_zs, axis=0)
    # Sort by ascending distance
    mean_delta_z = mean_delta_z[edge_sorting_idx]
    return mean_delta_z


def highlow_analysis(mean_qcs, corr_mats):
    """Perform high-low QC analysis.

    Parameters
    ----------
    mean_qcs : numpy.ndarray of shape (n_subjects,)
        QC measure (typically mean framewise displacement) across participants.
    corr_mats : numpy.ndarray of shape (n_subjects, n_edges)
        Z-transformed correlation coefficients for ROI-ROI pairs.
        n_edges is the *unique* ROI-to-ROI edges, not including self-self edges.
        These coefficients must be sorted according to ascending distance along the second axis.

    Returns
    -------
    hl_corr_diff : numpy.ndarray of shape (n_edges,)
        ROI-ROI pair difference scores.

    Notes
    -----
    The basic process for the high-low analysis is:
    1. Average QC values within each participant.
    2. Split the participants into high-QC and low-QC groups using a median split.
    3. Calculate the average z-transformed correlation coefficient for each group.
    4. Subtract the low group's value from the high group's value, for each ROI-ROI pair.
    """
    highgroup_idx = mean_qcs >= np.median(mean_qcs)
    lowgroup_idx = mean_qcs < np.median(mean_qcs)
    highgroup_mean_z = np.mean(corr_mats[highgroup_idx, :], axis=0)
    lowgroup_mean_z = np.mean(corr_mats[lowgroup_idx, :], axis=0)
    hl_corr_diff = highgroup_mean_z - lowgroup_mean_z
    return hl_corr_diff


def qcrsfc_analysis(mean_qcs, corr_mats):
    """Perform quality-control resting-state functional connectivity analysis.

    Parameters
    ----------
    mean_qcs : numpy.ndarray of shape (n_subjects,)
        QC measure (typically mean framewise displacement) across participants.
    corr_mats : numpy.ndarray of shape (n_subjects, n_edges)
        Z-transformed correlation coefficients for ROI-ROI pairs.
        n_edges is the *unique* ROI-to-ROI edges, not including self-self edges.
        These coefficients must be sorted according to ascending distance along the second axis.

    Returns
    -------
    qcrsfc_zs : numpy.ndarray of shape (n_edges,)
        Z-transformed correlation coefficients for ROI-ROI pairs.

    Notes
    -----
    The basic process for the QC:RSFC analysis is:
    1. Average QC values within each participant.
    2. Correlate the mean QC values with z-transformed correlation coefficients
       across participants, for each ROI-ROI pair.
    3. Z-transform the edge-wise correlation coefficients.
    """
    # Correlate each ROI pair's z-value against QC measure (usually FD) across subjects.
    qcrsfc_rs = fast_pearson(corr_mats.T, mean_qcs)
    qcrsfc_zs = np.arctanh(qcrsfc_rs)
    return qcrsfc_zs


def _scrubbing_null_iter(
    qc_values,
    ts_all,
    distances,
    qc_thresh,
    edge_sorting_idx,
    window,
    seed=0,
):
    perm_qcs = [np.random.RandomState(seed=seed).permutation(perm_qc) for perm_qc in qc_values]
    perm_mean_delta_zs = scrubbing_analysis(
        perm_qcs,
        ts_all,
        edge_sorting_idx,
        qc_thresh,
        perm=True,
    )
    perm_scrub_smoothing_curve = moving_average(perm_mean_delta_zs, window)
    perm_scrub_smoothing_curve, _ = average_across_distances(
        perm_scrub_smoothing_curve,
        distances,
    )
    return perm_scrub_smoothing_curve


def scrubbing_null_distribution(
    qc_values,
    ts_all,
    distances,
    qc_thresh,
    edge_sorting_idx,
    window=1000,
    n_iters=10000,
    n_jobs=1,
):
    """Generate null distribution smoothing curves for scrubbing analysis.

    Parameters
    ----------
    qc_values : list of n_subjects length containing numpy.ndarray of shape (n_timepoints,)
        QC time series for each participant.
    ts_all : list of n_subjects length containing numpy.ndarray of shape (n_timepoints, n_rois)
        ROI time series for each participant.
    distances : numpy.ndarray of shape (n_edges,)
        Distances for edges, already sorted in ascending order.
    qc_thresh : float
        QC threshold used to identify bad volumes (i.e., scrub).
    edge_sorting_idx : numpy.ndarray of shape (n_edges,)
        Sorting index of the flattened upper triangle (minus the diagonal) of the correlation
        matrix, in order of ascending distance.
    window : int, optional
        Sliding window to use for smoothing curve. Default is 1000.
    n_iters : int, optional
        Number of iterations with which to build the null distributions. Default is 10000.
    n_jobs : int, optional
        The number of CPUs to use to do the computation. -1 means 'all CPUs'. Default is 1.

    Returns
    -------
    perm_scrub_smoothing_curve : numpy.ndarray of shape (n_iters, n_unique_edge_distances)
        Smoothing curves for all permutations, to be used as a null distribution.
    """
    qc_values = [subj_qc_values.copy() for subj_qc_values in qc_values]

    scrub_null_smoothing_curves = Parallel(n_jobs=n_jobs)(
        delayed(_scrubbing_null_iter)(
            qc_values,
            ts_all,
            distances,
            qc_thresh,
            edge_sorting_idx,
            window,
            seed=seed,
        )
        for seed in range(n_iters)
    )

    scrub_null_smoothing_curves = np.vstack(scrub_null_smoothing_curves)

    return scrub_null_smoothing_curves


def _other_null_iter(mean_qcs, corr_mats, distances, window, seed=0):
    # Prep for QC:RSFC and high-low motion analyses
    perm_mean_qcs = np.random.RandomState(seed=seed).permutation(mean_qcs)

    # QC:RSFC analysis
    perm_qcrsfc_zs = qcrsfc_analysis(perm_mean_qcs, corr_mats)
    perm_qcrsfc_smoothing_curve = moving_average(perm_qcrsfc_zs, window)
    perm_qcrsfc_smoothing_curve, _ = average_across_distances(
        perm_qcrsfc_smoothing_curve,
        distances,
    )

    # High-low analysis
    perm_hl_diff = highlow_analysis(perm_mean_qcs, corr_mats)
    perm_hl_smoothing_curve = moving_average(perm_hl_diff, window)
    perm_hl_smoothing_curve, _ = average_across_distances(
        perm_hl_smoothing_curve,
        distances,
    )

    return perm_qcrsfc_smoothing_curve, perm_hl_smoothing_curve


def other_null_distributions(
    qc_values,
    corr_mats,
    distances,
    window=1000,
    n_iters=10000,
    n_jobs=1,
):
    """Generate null distribution smoothing curves for QC:RSFC and high-low analyses.

    Parameters
    ----------
    qc_values : list of n_subjects length containing numpy.ndarray of shape (n_timepoints,)
        QC time series for each participant.
    corr_mats : list of n_subjects length containing numpy.ndarray of shape (n_rois, n_rois)
        Z-transformed ROI-ROI correlation matrix for each participant.
    distances : numpy.ndarray of shape (n_edges,)
        Distances for edges, already sorted in ascending order.
    window : int, optional
        Sliding window to use for smoothing curve. Default is 1000.
    n_iters : int, optional
        Number of iterations with which to build the null distributions. Default is 10000.

    Returns
    -------
    perm_qcrsfc_smoothing_curve : numpy.ndarray of shape (n_iters, n_unique_edge_distances)
        Smoothing curves for all permutations, to be used as a null distribution.
    perm_hl_smoothing_curve : numpy.ndarray of shape (n_iters, n_unique_edge_distances)
        Smoothing curves for all permutations, to be used as a null distribution.
    """
    qc_values = [subj_qc_values.copy() for subj_qc_values in qc_values]
    mean_qcs = np.array([np.mean(subj_qc_values) for subj_qc_values in qc_values])

    results = Parallel(n_jobs=n_jobs)(
        delayed(_other_null_iter)(mean_qcs, corr_mats, distances, window, seed=seed)
        for seed in range(n_iters)
    )
    qcrsfc_null_smoothing_curves, hl_null_smoothing_curves = zip(*results)

    qcrsfc_null_smoothing_curves = np.vstack(qcrsfc_null_smoothing_curves)
    hl_null_smoothing_curves = np.vstack(hl_null_smoothing_curves)

    return qcrsfc_null_smoothing_curves, hl_null_smoothing_curves



"""Functions for the Earl filter."""
import math

import numpy as np
from scipy import signal

from .utils import get_fd_power


def respiration_iirnotch(TR_in_sec, bpm_min=18.582, bpm_max=25.7263):
    """Calculate filter parameters for respiration filter.

    Takes in the TR (optional min/max breaths-per-min, bpm_min, bpm_max).
    Returns the parameters for IIR Notch filter.
    """
    fs = 1.0 / TR_in_sec  # Sampling frequency (Hz)
    fn = fs / 2.0  # Nyquist frequency (Hz)

    # RR MIN
    rr_min = bpm_min / 60.0  # respiration rate minimum in Hz
    fa_min = abs(rr_min - math.floor((rr_min + fn) / fs) * fs)  # Aliased minimum frequency (Hz)
    w0_min = fa_min / fn  # Normalized minimum frequency

    # RR MAX
    rr_max = bpm_max / 60.0  # respiration rate maximum in Hz
    fa_max = abs(rr_max - math.floor((rr_max + fn) / fs) * fs)  # Aliased maximum frequency (Hz)
    w0_max = fa_max / fn  # Normalized maximum frequency

    # RR iirnotch filter
    w0 = np.mean([w0_min, w0_max])  # Mean normalized frequency
    bw = abs(w0_max - w0_min)  # Normalized bandwidth
    Q = w0 / bw  # Quality factor
    b, a = signal.iirnotch(w0, Q)  # Filter design

    return b, a


def filter_earl(motpars, t_r, radius=50):
    """Apply Earl filter to motion parameters.

    Parameters
    ----------
    motpars : (T, 6) array_like
        Raw motion parameters. First three columns are translations and last
        three are rotations. Rotations in degrees.
    t_r : :obj:`float`
        Repetition time in seconds.
    radius : :obj:`float`, optional
        Head radius for conversion of rotation units to distance at the edge of the brain.

    Returns
    -------
    motpars_filtered : (T, 6) array_like
    fd_filtered : (T,) array_like
    """
    # Create the filter
    b, a = respiration_iirnotch(t_r, bpm_min=18.582, bpm_max=25.7263)

    # Filter motion numbers
    filt_moco2 = signal.filtfilt(b, a, motpars, axis=0, padtype=None)
    filt_moco = signal.filtfilt(b, a, filt_moco2, axis=0, padtype=None)

    # Calculate FD values for the series
    fd = get_fd_power(filt_moco, order=["x", "y", "z", "r", "p", "ya"], unit="rad", radius=radius)

    return filt_moco, fd



#!/usr/bin/env python
# -*- coding: utf-8 -*-
""" ddmra setup script """


def main():
    """ Install entry-point """
    import os.path as op
    from inspect import currentframe, getfile
    from io import open

    from setuptools import find_packages, setup

    import versioneer

    ver_file = op.join("ddmra", "info.py")
    with open(ver_file) as f:
        exec(f.read())
    vars = locals()

    root_dir = op.dirname(op.abspath(getfile(currentframe())))
    cmdclass = versioneer.get_cmdclass()

    pkg_data = {
        "ddmra": [
            "tests/data/*",
        ]
    }

    setup(
        name=vars["PACKAGENAME"],
        version=vars["VERSION"],
        description=vars["DESCRIPTION"],
        long_description=vars["LONGDESC"],
        long_description_content_type="text/markdown",
        author=vars["AUTHOR"],
        author_email=vars["EMAIL"],
        maintainer=vars["MAINTAINER"],
        maintainer_email=vars["EMAIL"],
        url=vars["URL"],
        license=vars["LICENSE"],
        classifiers=vars["CLASSIFIERS"],
        download_url=vars["DOWNLOAD_URL"],
        # Dependencies handling
        python_requires=vars["PYTHON_REQUIRES"],
        install_requires=vars["REQUIRES"],
        tests_require=vars["TESTS_REQUIRES"],
        extras_require=vars["EXTRA_REQUIRES"],
        entry_points=vars["ENTRY_POINTS"],
        packages=find_packages(exclude=("tests",)),
        zip_safe=False,
        cmdclass=cmdclass,
    )


if __name__ == "__main__":
    main()



"""Perform distance-dependent motion-related artifact analyses."""
import logging
import os.path as op
from os import makedirs

import numpy as np
import pandas as pd
from nilearn import datasets, input_data
from scipy.spatial.distance import pdist, squareform

from . import analysis, plotting, utils

LGR = logging.getLogger("workflows")


def run_analyses(
    files,
    qc,
    out_dir=".",
    confounds=None,
    n_iters=10000,
    n_jobs=1,
    qc_thresh=0.2,
    window=1000,
):
    """Run scrubbing, high-low motion, and QCRSFC analyses.

    Parameters
    ----------
    files : (N,) list of nifti files
        List of 4D (X x Y x Z x T) images in MNI space.
    qc : (N,) list of array_like
        List of 1D (T) numpy arrays with QC metric values per img (e.g., FD or respiration).
    out_dir : str, optional
        Output directory. Default is current directory.
    confounds : None or (N,) list of array-like, optional
        List of 2D (T) numpy arrays with confounds per img.
        Default is None (no confounds are removed).
    n_iters : int, optional
        Number of iterations to run to generate null distributions. Default is 10000.
    n_jobs : int, optional
        The number of CPUs to use to do the computation. -1 means 'all CPUs'. Default is 1.
    qc_thresh : float, optional
        Threshold for QC metric used in scrubbing analysis. Default is 0.2 (for FD).
    window : int, optional
        Number of units (pairs of ROIs) to include when averaging to generate smoothing curve.
        Default is 1000.

    Notes
    -----
    This function writes out several files to out_dir:
    - ``analysis_values.tsv.gz``: Raw analysis values for analyses.
        Has four columns: distance, qcrsfc, scrubbing, and highlow.
    - ``smoothing_curves.tsv.gz``: Smoothing curve information for analyses.
        Has four columns: distance, qcrsfc, scrubbing, and highlow.
    - ``null_smoothing_curves.npz``:
        Null smoothing curves from each analysis.
        Contains three 2D arrays, where number of columns is same
        size and order as distance column in ``smoothing_curves.tsv.gz``
        and number of rows is number of iterations for permutation analysis.
        The three arrays' keys are 'qcrsfc', 'highlow', and 'scrubbing'.
    - ``[analysis]_analysis.png``: Figure for each analysis.
    """
    makedirs(out_dir, exist_ok=True)

    # create LGR with 'spam_application'
    LGR.setLevel(logging.DEBUG)
    # create file handler which logs even debug messages
    fh = logging.FileHandler(op.join(out_dir, "log.tsv"))
    fh.setLevel(logging.DEBUG)
    # create formatter and add it to the handlers
    formatter = logging.Formatter("%(asctime)s\t%(name)-12s\t%(levelname)-8s\t%(message)s")
    fh.setFormatter(formatter)
    # add the handlers to the LGR
    LGR.addHandler(fh)

    LGR.info("Preallocating matrices")
    n_subjects = len(files)

    # Load atlas and associated masker
    atlas = datasets.fetch_coords_power_2011()
    coords = np.vstack((atlas.rois["x"], atlas.rois["y"], atlas.rois["z"])).T
    n_rois = coords.shape[0]
    triu_idx = np.triu_indices(n_rois, k=1)
    distances = squareform(pdist(coords))
    distances = distances[triu_idx]

    # Sorting index for distances
    edge_sorting_idx = distances.argsort()
    distances = distances[edge_sorting_idx]

    LGR.info("Creating masker")
    spheres_masker = input_data.NiftiSpheresMasker(
        seeds=coords,
        radius=5.0,
        t_r=None,
        smoothing_fwhm=4.0,
        detrend=False,
        standardize=False,
        low_pass=None,
        high_pass=None,
    )

    # prep for qcrsfc and high-low motion analyses
    mean_qc = np.array([np.mean(subj_qc) for subj_qc in qc])
    z_corr_mats = np.zeros((n_subjects, distances.size))

    # Get correlation matrices
    ts_all = []
    LGR.info("Building correlation matrices")
    if confounds:
        LGR.info("Regressing confounds out of data.")

    for i_subj in range(n_subjects):
        if confounds:
            raw_ts = spheres_masker.fit_transform(files[i_subj], confounds=confounds[i_subj]).T
        else:
            raw_ts = spheres_masker.fit_transform(files[i_subj]).T

        assert raw_ts.shape[0] == n_rois

        ts_all.append(raw_ts)
        raw_corrs = np.corrcoef(raw_ts)
        raw_corrs = raw_corrs[triu_idx]
        raw_corrs = raw_corrs[edge_sorting_idx]  # Sort from close to far ROI pairs
        z_corr_mats[i_subj, :] = np.arctanh(raw_corrs)

    del (raw_corrs, raw_ts, spheres_masker, atlas, coords)

    analysis_values = pd.DataFrame(columns=["qcrsfc", "highlow", "scrubbing"], index=distances)
    analysis_values.index.name = "distance"

    # QC:RSFC r analysis
    LGR.info("Performing QC:RSFC analysis")
    qcrsfc_values = analysis.qcrsfc_analysis(mean_qc, z_corr_mats)
    analysis_values["qcrsfc"] = qcrsfc_values
    qcrsfc_smoothing_curve = utils.moving_average(qcrsfc_values, window)
    qcrsfc_smoothing_curve, smoothing_curve_distances = utils.average_across_distances(
        qcrsfc_smoothing_curve,
        distances,
    )

    # Quick interlude to create the smoothing_curves DataFrame
    smoothing_curves = pd.DataFrame(
        columns=["qcrsfc", "highlow", "scrubbing"],
        index=smoothing_curve_distances,
    )
    smoothing_curves.index.name = "distance"

    smoothing_curves.loc[smoothing_curve_distances, "qcrsfc"] = qcrsfc_smoothing_curve
    del qcrsfc_values, qcrsfc_smoothing_curve

    # High-low motion analysis
    LGR.info("Performing high-low motion analysis")
    highlow_values = analysis.highlow_analysis(mean_qc, z_corr_mats)
    analysis_values["highlow"] = highlow_values
    hl_smoothing_curve = utils.moving_average(highlow_values, window)
    hl_smoothing_curve, smoothing_curve_distances = utils.average_across_distances(
        hl_smoothing_curve,
        distances,
    )
    smoothing_curves.loc[smoothing_curve_distances, "highlow"] = hl_smoothing_curve
    del highlow_values, hl_smoothing_curve

    # Scrubbing analysis
    LGR.info("Performing scrubbing analysis")
    scrub_values = analysis.scrubbing_analysis(qc, ts_all, edge_sorting_idx, qc_thresh, perm=False)
    analysis_values["scrubbing"] = scrub_values
    scrub_smoothing_curve = utils.moving_average(scrub_values, window)
    scrub_smoothing_curve, smoothing_curve_distances = utils.average_across_distances(
        scrub_smoothing_curve,
        distances,
    )
    smoothing_curves.loc[smoothing_curve_distances, "scrubbing"] = scrub_smoothing_curve
    del scrub_values, scrub_smoothing_curve

    analysis_values.reset_index(inplace=True)
    smoothing_curves.reset_index(inplace=True)

    analysis_values.to_csv(
        op.join(out_dir, "analysis_values.tsv.gz"),
        sep="\t",
        line_terminator="\n",
        index=False,
    )
    smoothing_curves.to_csv(
        op.join(out_dir, "smoothing_curves.tsv.gz"),
        sep="\t",
        line_terminator="\n",
        index=False,
    )

    # Null distributions
    LGR.info("Building null distributions with permutations")
    qcrsfc_null_smoothing_curves, hl_null_smoothing_curves = analysis.other_null_distributions(
        qc,
        z_corr_mats,
        distances,
        window=window,
        n_iters=n_iters,
        n_jobs=n_jobs,
    )
    scrub_null_smoothing_curves = analysis.scrubbing_null_distribution(
        qc,
        ts_all,
        distances,
        qc_thresh,
        edge_sorting_idx,
        window=window,
        n_iters=n_iters,
        n_jobs=n_jobs,
    )

    np.savez_compressed(
        op.join(out_dir, "null_smoothing_curves.npz"),
        qcrsfc=qcrsfc_null_smoothing_curves,
        highlow=hl_null_smoothing_curves,
        scrubbing=scrub_null_smoothing_curves,
    )

    del qcrsfc_null_smoothing_curves, hl_null_smoothing_curves, scrub_null_smoothing_curves

    plotting.plot_results(out_dir)

    LGR.info("Workflow completed")
